#include "hydra-mod.h"

#include "ntlm.h"
#ifdef LIBOPENSSLNEW
  #include <openssl/md5.h>
#endif

extern char *HYDRA_EXIT;
char *buf;
char *webtarget = NULL;
char *slash = "/";
int webport;

#define AUTH_BASIC 0
#define AUTH_DIGESTMD5 1
#define AUTH_NTLM 2

int http_auth_mechanism=AUTH_BASIC;

int start_http(int s, char *ip, int port, unsigned char options, char *miscptr, FILE * fp, char *type) {
  char *empty = "";
  char *login, *pass, buffer[500], buffer2[500];
  char *header = "";            /* XXX TODO */
  char *ptr;

  if (strlen(login = hydra_get_next_login()) == 0)
    login = empty;
  if (strlen(pass = hydra_get_next_password()) == 0)
    pass = empty;

  if (verbose)
	hydra_report(stderr, "Testing login: %s with password: %s\n", login ,pass);

   switch (http_auth_mechanism){
    case AUTH_BASIC: 
      sprintf(buffer2, "%.50s:%.50s", login, pass);
      hydra_tobase64((unsigned char *) buffer2, strlen(buffer2), sizeof(buffer2));

      /* again: no snprintf to be portable. dont worry, buffer cant overflow */
      if (use_proxy == 1 && proxy_authentication != NULL)
	sprintf(buffer, "%s http://%s:%d%.250s HTTP/1.0\r\nHost: %s\r\nAuthorization: Basic %s\r\nProxy-Authorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
        	type, webtarget, webport, miscptr, webtarget, buffer2, proxy_authentication, header);
      else {
	if (use_proxy == 1)
	  sprintf(buffer, "%s http://%s:%d%.250s HTTP/1.0\r\nHost: %s\r\nAuthorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
        	type, webtarget, webport, miscptr, webtarget, buffer2, header);
	else
	  sprintf(buffer, "%s %.250s HTTP/1.0\r\nHost: %s\r\nAuthorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
        	type, miscptr, webtarget, buffer2, header);
      }
      if (verbose)
        hydra_report(stderr, "C1:%s\n", buffer);
      break;

#ifdef LIBOPENSSLNEW
    case AUTH_DIGESTMD5: {
      char *pbuffer;
      int array_size=10;
      char *array[array_size];
      char buffer3[500], nonce[200], realm[50], algo[10];
      int i, ind=0, lastpos=0, currentpos=0, intq=0;
      MD5_CTX md5c;
      unsigned char response[MD5_DIGEST_LENGTH];

      pbuffer=strstr(buf, "WWW-Authenticate: Digest ");
      strncpy(buffer, pbuffer+strlen("WWW-Authenticate: Digest "), sizeof(buffer));
      buffer[sizeof(buffer)-1]='\0';
      pbuffer=buffer;

      do{
	currentpos++;
	if (pbuffer[0] == '"') {
	  if (intq == 0)
	    intq=1;
	  else {
	    intq=0;
	  }
	}
	if ((pbuffer[0] == ',') && (intq==0)) {
	  array[ind]=malloc(currentpos);
	  strncpy(array[ind],buffer+lastpos,currentpos-1);
	  array[ind][currentpos-1]='\0';
	  ind++;
	  lastpos+=currentpos;
	  currentpos=0;
	}
	pbuffer++;
      } while ((pbuffer[0] != '\n') && (ind<array_size));

      //save the latest one	
      array[ind]=malloc(currentpos+1);
      strncpy(array[ind],buffer+lastpos,currentpos+1);
      array[ind][currentpos]='\0';
      ind++;

      for(i=0;i<ind;i++) {

	if (strstr(array[i], "nonce=") != NULL) {
	  strcpy(nonce, strstr(array[i], "nonce=")+strlen("nonce=")+1);
	  nonce[strlen(nonce)-1]='\0';
	}
	if (strstr(array[i], "realm=") != NULL) {
	  strcpy(realm, strstr(array[i], "realm=")+strlen("realm=")+1);
	  realm[strlen(realm)-1]='\0';
	}
	if (strstr(array[i], "qop=") != NULL) {
          if ((strstr(array[i], "\"auth\"") == NULL) && (strstr(array[i], "\"auth,") == NULL) && (strstr(array[i], ",auth\"") == NULL)) {
    	    hydra_report(stderr, "Error: HTTP DIGEST-MD5 quality of protection is not supported\n");
	    return 3;
	  }
       }

       if (strstr(array[i], "algorithm=") != NULL) {
	  strcpy(algo, strstr(array[i], "algorithm=")+strlen("algorithm="));
	  if (algo[strlen(algo)-1]=='"')
	    algo[strlen(algo)-1]='\0';
	  if (strstr(algo, "MD5") == NULL) {
	    hydra_report(stderr, "Error: HTTP DIGEST-MD5 algorithm is not supported\n");
	    return 3;
	  }
       }
       free(array[i]);
      }

      //compute ha1
      //support for algo = MD5
      sprintf(buffer, "%s:%s:%s", login, realm, pass);
      MD5_Init(&md5c);
      MD5_Update(&md5c, buffer, strlen(buffer));
      MD5_Final(response, &md5c);

      //for MD5-sess
      if (strstr(algo, "MD5-sess")) {
	sprintf(buffer, "%s:%s:%s", response ,nonce, "hydra");
	MD5_Init(&md5c);
	MD5_Update(&md5c, buffer, strlen(buffer));
	MD5_Final(response, &md5c);
      }

      pbuffer = buffer3;
      for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	sprintf(pbuffer, "%02x", response[i]);
	pbuffer +=2;
      }

      //compute ha2
      sprintf(buffer, "%s:%s",type, miscptr);
      MD5_Init(&md5c);
      MD5_Update(&md5c, buffer, strlen(buffer));
      MD5_Final(response, &md5c);

      pbuffer = buffer2;
      for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	sprintf(pbuffer, "%02x", response[i]);
	pbuffer +=2;
      }

      //compute response
      sprintf(buffer, "%s:%s:%s:%s:%s", nonce,"00000001", "hydra", "auth", buffer2);
      MD5_Init(&md5c);
      MD5_Update(&md5c, buffer3, strlen(buffer3));
      MD5_Update(&md5c, ":", 1);
      MD5_Update(&md5c, buffer, strlen(buffer));
      MD5_Final(response, &md5c);
      pbuffer = buffer;
      for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	sprintf(pbuffer, "%02x", response[i]);
	pbuffer +=2;
      }

      //create the auth response
      if (use_proxy == 1 && proxy_authentication != NULL)
        sprintf(buffer2, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: Digest username=\"%s\", realm=\"%s\", response=\"%s\", nonce=\"%s\", cnonce=\"hydra\", nc=00000001, algorithm=%s, qop=auth, uri=\"%s\"\r\nProxy-Authorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, login, realm, buffer, nonce, algo, miscptr, proxy_authentication, header);
      else {
	if (use_proxy == 1)
          sprintf(buffer2, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: Digest username=\"%s\", realm=\"%s\", response=\"%s\", nonce=\"%s\", cnonce=\"hydra\", nc=00000001, algorithm=%s, qop=auth, uri=\"%s\"\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, login, realm, buffer, nonce, algo, miscptr, header);
	else
          sprintf(buffer2, "%s %s HTTP/1.0\r\nHost: %s\r\nAuthorization: Digest username=\"%s\", realm=\"%s\", response=\"%s\", nonce=\"%s\", cnonce=\"hydra\", nc=00000001, algorithm=%s, qop=auth, uri=\"%s\"\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, miscptr, webtarget, login, realm, buffer, nonce, algo, miscptr, header);
      }

      if (verbose)
	hydra_report(stderr, "C:%s\n", buffer2);
      strcpy(buffer, buffer2);
    }
    break;
#endif

    case AUTH_NTLM: {  
      unsigned char buf1[4096];
      unsigned char buf2[4096];

      //send auth and receive challenge
      //send auth request: let the server send it's own hostname and domainname
      buildAuthRequest((tSmbNtlmAuthRequest*)buf2,0,NULL,NULL);
      to64frombits(buf1, buf2, SmbLength((tSmbNtlmAuthRequest*)buf2));

      /* to be portable, no snprintf, buffer is big enough so it cant overflow */
      //send the first..
      if (use_proxy == 1 && proxy_authentication != NULL)
        sprintf(buffer, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nProxy-Authorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, buf1, proxy_authentication, header);
      else {
	if (use_proxy == 1)
          sprintf(buffer, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, buf1, header);
	else
	  sprintf(buffer, "%s %s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n", 
		type, miscptr, webtarget, buf1, header);
      }

      if (hydra_send(s, buffer, strlen(buffer), 0) < 0)
	return 1;

      //receive challenge
      buf = hydra_receive_line(s);
      while (strstr(buf, "WWW-Authenticate: NTLM ") == NULL && buf != NULL)
	buf = hydra_receive_line(s);

      //recover challenge
      from64tobits((char*)buf1, buf+23);

      //Send response
      buildAuthResponse((tSmbNtlmAuthChallenge*)buf1,(tSmbNtlmAuthResponse*)buf2,0,login,pass,NULL,NULL);
      to64frombits(buf1, buf2, SmbLength((tSmbNtlmAuthResponse*)buf2));
      
      //create the auth response
      if (use_proxy == 1 && proxy_authentication != NULL)
        sprintf(buffer, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nProxy-Authorization: Basic %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, buf1, proxy_authentication, header);
      else {
	if (use_proxy == 1)
          sprintf(buffer, "%s http://%s:%d%s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n",
		type, webtarget, webport, miscptr, webtarget, buf1, header);
	else
      	  sprintf(buffer, "%s %s HTTP/1.0\r\nHost: %s\r\nAuthorization: NTLM %s\r\nUser-Agent: Mozilla/4.0 (Hydra)\r\n%s\r\n", 
		type, miscptr, webtarget, buf1, header);
      }

      if (verbose)
	 hydra_report(stderr, "C:%s\n", buffer);
    }
    break;
  }

  if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
    return 1;
  }

  buf = hydra_receive_line(s);
  while (strstr(buf, "HTTP/1.") == NULL && buf != NULL)
    buf = hydra_receive_line(s);

  if (buf == NULL)
    return 1;

  if (verbose)
	hydra_report(stderr, "S:%s\n", buf);

  ptr = ((char *) index(buf, ' ')) + 1;
  if (ptr != NULL && (*ptr == '2' || *ptr == '3' || strncmp(ptr, "403", 3) == 0 || strncmp(ptr, "404", 3) == 0)) {
    hydra_report_found_host(port, ip, "www", fp);
    hydra_completed_pair_found();
  } else {
    if (ptr != NULL && *ptr != '4')
      printf("Unusual return code: %c for %s:%s\n", (char) *(index(buf, ' ') + 1), login, pass);

    //the first authentication type failed, check the type from server header
    if ((strstr(buf, "WWW-Authenticate: Basic") == NULL) && (http_auth_mechanism==AUTH_BASIC)) {
	  //seems the auth supported is not Basic shceme so testing further
	  int find_auth=0;
	  if (strstr(buf, "WWW-Authenticate: NTLM") != NULL) {
		  http_auth_mechanism=AUTH_NTLM;
		  find_auth=1;
	  }
#ifdef LIBOPENSSLNEW
	  if (strstr(buf, "WWW-Authenticate: Digest") != NULL) {
		  http_auth_mechanism=AUTH_DIGESTMD5;
		  find_auth=1;
	  }
#endif

	  if (find_auth) {
		  return 1;
	  }
    }
    hydra_completed_pair();
  }
  free(buf);
  if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
    return 3;
  return 1;

}

void
service_http(char *ip, int sp, unsigned char options, char *miscptr, FILE * fp, int port, char *type)
{
  int run = 1, next_run, sock = -1;
  int myport = PORT_HTTP, mysslport = PORT_HTTP_SSL;
  char *ptr, *ptr2;

  hydra_register_socket(sp);
  if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
    return;

  if ((webtarget = strstr(miscptr, "://")) != NULL) {
    webtarget += strlen("://");
    if ((ptr2 = index(webtarget, ':')) != NULL) { /* step over port if present */
      *ptr2 = 0;
      ptr2++;
      ptr = ptr2;
      if (*ptr == '/' || (ptr = index(ptr2, '/')) != NULL)
        miscptr = ptr;
      else
        miscptr = slash; /* to make things easier to user */
    } else if ((ptr2 = index(webtarget, '/')) != NULL) {
      miscptr = malloc(strlen(ptr2) + 1);
      strcpy(miscptr, ptr2);
      *ptr2 = 0;
    } else
      webtarget = NULL;
  }
  if (cmdlinetarget != NULL && webtarget == NULL)
    webtarget = cmdlinetarget;
  else
    if (webtarget == NULL && cmdlinetarget == NULL)
      webtarget = hydra_address2string(ip);
  if (port != 0)
    webport = port;
  else if ((options & OPTION_SSL) == 0)
    webport = myport;
  else
    webport = mysslport;

  while (1) {
    next_run = 0;
    switch (run) {
    case 1:                    /* connect and service init function */
      {

        if (sock >= 0)
          sock = hydra_disconnect(sock);
//        usleep(275000);
        if ((options & OPTION_SSL) == 0) {
          if (port != 0)
            myport = port;
          sock = hydra_connect_tcp(ip, myport);
          port = myport;
        } else {
          if (port != 0)
            mysslport = port;
          sock = hydra_connect_ssl(ip, mysslport);
          port = mysslport;
        }
        if (sock < 0) {
          fprintf(stderr, "Error: Child with pid %d terminating, can not connect\n", (int) getpid());
          hydra_child_exit(1);
        }
        next_run = 2;
        break;
      }
    case 2:                    /* run the cracking function */
      next_run = start_http(sock, ip, port, options, miscptr, fp, type);
      break;
    case 3:                    /* clean exit */
      if (sock >= 0)
        sock = hydra_disconnect(sock);
      hydra_child_exit(0);
      return;
    default:
      fprintf(stderr, "Caught unknown return code, exiting!\n");
      hydra_child_exit(0);
      exit(-1);
    }
    run = next_run;
  }
}

void
service_http_get(char *ip, int sp, unsigned char options, char *miscptr, FILE * fp, int port)
{
  service_http(ip, sp, options, miscptr, fp, port, "GET");
}

void
service_http_head(char *ip, int sp, unsigned char options, char *miscptr, FILE * fp, int port)
{
  service_http(ip, sp, options, miscptr, fp, port, "HEAD");
}
