#include "hydra-mod.h"
#include "ntlm.h"
#ifdef LIBOPENSSLNEW
  #include <openssl/md5.h>
#endif

#define AUTH_LOGIN 1
#define AUTH_PLAIN 2
#define AUTH_CRAMMD5 3
#define AUTH_DIGESTMD5 4
#define AUTH_NTLM 5


extern char *HYDRA_EXIT;
int from64tobits(char *out, const char *in);
extern void to64frombits(unsigned char *out, const unsigned char *in, int inlen);

char *buf;
int smtp_auth_mechanism=AUTH_LOGIN;

int
start_smtpauth(int s, char *ip, int port, unsigned char options,
	       char *miscptr, FILE * fp)
{
  char *empty = "";
  char *login, *pass, buffer[500], buffer2[500];

  if (strlen(login = hydra_get_next_login()) == 0)
    login = empty;
  if (strlen(pass = hydra_get_next_password()) == 0)
    pass = empty;

  while (hydra_data_ready(s) > 0) {
    if ((buf = hydra_receive_line(s)) == NULL)
      return (1);
    free(buf);
  }
  if (verbose)
	hydra_report(stderr, "Testing login: %s with password: %s\n", login ,pass);

  switch (smtp_auth_mechanism) {

     case AUTH_PLAIN:
	sprintf(buffer, "AUTH PLAIN\r\n");
	if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
	  return 1;
	}
	if ((buf = hydra_receive_line(s)) == NULL)
	  return 1;
	if (strstr(buf, "334") == NULL) {
	  hydra_report(stderr, "Error: SMTP PLAIN AUTH error: %s\n", buf);
	  free(buf);
	  return 3;
	}
	free(buf);

	memset(buffer, 0, sizeof(buffer));
	if (2*strlen(login) + 3 + strlen(pass) < 250) {
	  strcpy(buffer, login);
	  strcpy(buffer +  strlen(login) + 1, login);
	  strcpy(buffer + 2*strlen(login) + 2, pass);
 	  hydra_tobase64((unsigned char *) buffer, strlen(login)*2+strlen(pass)+2, sizeof(buffer));
	}
	break;

#ifdef LIBOPENSSLNEW
    case AUTH_CRAMMD5: {
	MD5_CTX c;
	int i=0;
	unsigned char md5_raw[MD5_DIGEST_LENGTH];

	char ipad[64];
	char opad[64];

	memset(ipad, 0, sizeof(ipad));
	memset(opad, 0, sizeof(opad));

	sprintf(buffer, "AUTH CRAM-MD5\r\n");
	if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
	  return 1;
	}
	//get the one-time BASE64 encoded challenge
	if ((buf = hydra_receive_line(s)) == NULL)
	  return 1;
	if (strstr(buf, "334") == NULL) {
	  hydra_report(stderr, "Error: SMTP CRAM-MD5 AUTH error: %s\n", buf);
	  free(buf);
	  return 3;
	}
	
	memset(buffer, 0, sizeof(buffer));
	from64tobits((char*)buffer, buf+4);
	free(buf);

	if (strlen(pass) >= 64) {
		MD5_Init(&c);
		MD5_Update(&c, pass, strlen(pass));
		MD5_Final(md5_raw, &c);
		memcpy(ipad, md5_raw, 16);
		memcpy(opad, md5_raw, 16);
	} else {
		strcpy(ipad, pass); // safe
		strcpy(opad, pass); // safe
	}

	for (i = 0; i < 64; i++) {
		ipad[i] ^= 0x36;
        	opad[i] ^= 0x5c;
	}

	MD5_Init(&c);
	MD5_Update(&c, ipad, 64);
	MD5_Update(&c, buffer, strlen(buffer));
	MD5_Final(md5_raw, &c);

	MD5_Init(&c);
	MD5_Update(&c, opad, 64);
	MD5_Update(&c, md5_raw, 16);
	MD5_Final(md5_raw, &c);	

	char *pbuffer = buffer2;
	for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
		sprintf(pbuffer, "%02x", md5_raw[i]);
		pbuffer +=2;
	}

	sprintf(buffer, "%s %.250s", login, buffer2);
	hydra_tobase64((unsigned char *) buffer, strlen(buffer), sizeof(buffer));
	sprintf(buffer, "%.250s\r\n", buffer);
	}
	break;

   case AUTH_DIGESTMD5: {
	MD5_CTX md5c;
	unsigned char response[MD5_DIGEST_LENGTH];
	char *pbuffer;
	int array_size=10;
	char *array[array_size];
	char buffer3[500], nonce[200], realm[50];
	int i, ind=0, lastpos=0, currentpos=0, intq=0;

	sprintf(buffer, "AUTH DIGEST-MD5\r\n");

	if (hydra_send(s, buffer, strlen(buffer), 0) < 0) return 1;
	//receive
	if ((buf = hydra_receive_line(s)) == NULL)
	  return 1;
	if (strstr(buf, "334") == NULL) {
	  hydra_report(stderr, "Error: SMTP DIGEST-MD5 AUTH error: %s\n", buf);
	  free(buf);
	  return 3;
	}
	memset(buffer, 0,sizeof(buffer));
	from64tobits((char*)buffer, buf+4);

	if (verbose)
		hydra_report(stderr, "DEBUG S: %s\n", buffer);

	pbuffer=buffer;

	do{
	  currentpos++;
	  if (pbuffer[0] == '"') {
	    if (intq == 0)
	      intq=1;
	    else {
	      intq=0;
	    }
	  }

	  if ((pbuffer[0] == ',') && (intq==0)) {
	    array[ind]=malloc(currentpos);
	    strncpy(array[ind],buffer+lastpos,currentpos-1);
	    array[ind][currentpos-1]='\0';
	    ind++;
	    lastpos+=currentpos;
	    currentpos=0;
	  }
	  pbuffer++;
	} while ((pbuffer[0] != '\0') && (ind<array_size));

	//save the latest one	
	array[ind]=malloc(currentpos+1);
	strncpy(array[ind],buffer+lastpos,currentpos+1);
	array[ind][currentpos]='\0';
	ind++;

	for(i=0;i<ind;i++) {
	  if (strstr(array[i], "nonce=") != NULL) {
	    strcpy(nonce, array[i]+strlen("nonce=")+1);
	    nonce[strlen(nonce)-1]='\0';
	  }
	  if (strstr(array[i], "realm=") != NULL) {
	    strcpy(realm, array[i]+strlen("realm=")+1);
	    realm[strlen(realm)-1]='\0';
	  }
	  if (strstr(array[i], "qop=") != NULL) {
            if ((strstr(array[i], "\"auth\"") == NULL) && (strstr(array[i], "\"auth,") == NULL) && (strstr(array[i], ",auth\"") == NULL)) {
    	     hydra_report(stderr, "Error: SMTP DIGEST-MD5 quality of protection is not supported\n");
	     return 3;
	    }
	 }
	 free(array[i]);
	}

	//compute ha1
	sprintf(buffer, "%s:%s:%s", login, realm, pass);
	MD5_Init(&md5c);
	MD5_Update(&md5c, buffer, strlen(buffer));
	MD5_Final(response, &md5c);
	sprintf(buffer, "%s:%s:%s", response ,nonce, "hydra");
	MD5_Init(&md5c);
	MD5_Update(&md5c, buffer, strlen(buffer));
	MD5_Final(response, &md5c);
	pbuffer = buffer3;
	for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	  sprintf(pbuffer, "%02x", response[i]);
	  pbuffer +=2;
	}

	//compute ha2
	sprintf(buffer, "AUTHENTICATE:smtp/%s", hydra_address2string(ip));
	MD5_Init(&md5c);
	MD5_Update(&md5c, buffer, strlen(buffer));
	MD5_Final(response, &md5c);

	pbuffer = buffer2;
	for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	  sprintf(pbuffer, "%02x", response[i]);
	  pbuffer +=2;
	}

	//compute response
	sprintf(buffer, "%s:%s:%s:%s:%s", nonce,"00000001", "hydra", "auth", buffer2);
	MD5_Init(&md5c);
	MD5_Update(&md5c, buffer3, strlen(buffer3));
	MD5_Update(&md5c, ":", 1);
	MD5_Update(&md5c, buffer, strlen(buffer));
	MD5_Final(response, &md5c);
	pbuffer = buffer;
	for(i = 0; i < MD5_DIGEST_LENGTH; i++) {
	  sprintf(pbuffer, "%02x", response[i]);
	  pbuffer +=2;
	}

	sprintf(buffer2, "username=\"%s\",realm=\"%s\",nonce=\"%s\",cnonce=\"hydra\",nc=00000001,qop=\"auth\",digest-uri=\"smtp/%s\",response=%s", login, realm, nonce, hydra_address2string(ip),buffer);
	if (verbose)	
	  hydra_report(stderr, "DEBUG C: %s\n", buffer2);
	hydra_tobase64((unsigned char *) buffer2, strlen(buffer2), sizeof(buffer2));
	sprintf(buffer, "%s\r\n", buffer2);
	}
	break;
#endif

   case AUTH_NTLM: {
	unsigned char buf1[4096];
	unsigned char buf2[4096];

	//send auth and receive challenge
	buildAuthRequest((tSmbNtlmAuthRequest*)buf2,0,NULL,NULL);
	to64frombits(buf1, buf2, SmbLength((tSmbNtlmAuthRequest*)buf2));
	sprintf(buffer, "AUTH NTLM %s\r\n", buf1);
	if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
	  return 1;
	}
	if ((buf = hydra_receive_line(s)) == NULL)
	  return 1;
	if (strstr(buf, "334") == NULL) {
	  hydra_report(stderr, "Error: SMTP NTLM AUTH error: %s\n", buf);
	  free(buf);
	  return 3;
	}
	//recover challenge
	from64tobits((char*)buf1, buf+4);
	free(buf);

	buildAuthResponse((tSmbNtlmAuthChallenge*)buf1,(tSmbNtlmAuthResponse*)buf2,0,login,pass,NULL,NULL);
	to64frombits(buf1, buf2, SmbLength((tSmbNtlmAuthResponse*)buf2));
	sprintf(buffer, "%s\r\n", buf1);
	}
	break;

   default:
      /* by default trying AUTH LOGIN */
      sprintf(buffer, "AUTH LOGIN\r\n");
      if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
	return 1;
      }
      if ((buf = hydra_receive_line(s)) == NULL)
	return 1;
      if (strstr(buf, "334") == NULL) {
	hydra_report(stderr, "Error: SMTP LOGIN AUTH error: %s\n", buf);
	free(buf);
	return 3;
      }
      free(buf);
      sprintf(buffer2, "%.250s", login);
      hydra_tobase64((unsigned char *) buffer2, strlen(buffer2), sizeof(buffer2));
      sprintf(buffer, "%.250s\r\n", buffer2);

      if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
	return 1;
      }
      if ((buf = hydra_receive_line(s)) == NULL)
	return (1);
      if (strstr(buf, "334") == NULL) {
	hydra_report(stderr, "Error: SMTP LOGIN AUTH error: %s\n", buf);
	free(buf);
	return (3);
      }
      free(buf);

      sprintf(buffer2, "%.250s", pass);
      hydra_tobase64((unsigned char *) buffer2, strlen(buffer2), sizeof(buffer2));
      sprintf(buffer, "%.250s\r\n", buffer2);    
  }

  if (hydra_send(s, buffer, strlen(buffer), 0) < 0) {
       return 1;
  }
  if ((buf = hydra_receive_line(s)) == NULL)
    return (1);

#ifdef LIBOPENSSLNEW
  if (smtp_auth_mechanism == AUTH_DIGESTMD5) {
    if (strstr(buf, "334") != NULL) {
      memset(buffer2, 0,sizeof(buffer2));
      from64tobits((char*)buffer2, buf+4);
      if (strstr(buffer2, "rspauth=") != NULL) {
        hydra_report_found_host(port, ip, "smtpauth", fp);
        hydra_completed_pair_found();
        free(buf);
        if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
	  return 3;
        return 1;
      }
    }
  } else
#endif
  {
    if (strstr(buf, "235") != NULL) {
      hydra_report_found_host(port, ip, "smtpauth", fp);
      hydra_completed_pair_found();
      free(buf);
      if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
	return 3;
      return 1;
    }
  }
  free(buf);
  hydra_completed_pair();
  if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
    return 3;

  return 2;
}

void
service_smtpauth(char *ip, int sp, unsigned char options,
		 char *miscptr, FILE * fp, int port)
{
  int run = 1, next_run, sock = -1, resp = 0, i = 0;
  int myport = PORT_SMTPAUTH, mysslport = PORT_SMTPAUTH_SSL;
  char *ptr;

  char *buffer1 = "EHLO hydra\r\n";
  char *buffer2 = "HELO hydra\r\n";

  hydra_register_socket(sp);
  if (memcmp(hydra_get_next_pair(), &HYDRA_EXIT, sizeof(HYDRA_EXIT)) == 0)
    return;
  while (1) {
    switch (run) {
    case 1:			/* connect and service init function */
      if (sock >= 0)
	sock = hydra_disconnect(sock);
      if ((options & OPTION_SSL) == 0) {
	if (port != 0)
	  myport = port;
	sock = hydra_connect_tcp(ip, myport);
	port = myport;
      } else {
	if (port != 0)
	  mysslport = port;
	sock = hydra_connect_ssl(ip, mysslport);
	port = myport;
      }
      if (sock < 0) {
	hydra_report(stderr,
		     "Error: Child with pid %d terminating, can not connect\n",
		     (int) getpid());
	hydra_child_exit(1);
      }

      /* receive initial header */
      if ((buf = hydra_receive_line(sock)) == NULL)
	exit(-1);
      if (strstr(buf, "220") == NULL) {
	hydra_report(stderr, "Warning: SMTP does not allow to connect: %s\n",
		     buf);
	hydra_child_exit(2);
	exit(-1);
      }
      while (strstr(buf, "220 ") == NULL) {
        free(buf);
        buf = hydra_receive_line(sock);
      }
      free(buf);

      /* send ehlo and receive/ignore reply */
      if (hydra_send(sock, buffer1, strlen(buffer1), 0) < 0)
	exit(-1);
      resp = 0;
      do {
        ptr = buf = hydra_receive_line(sock);
        if (buf != NULL) {
          if (isdigit((int)buf[0]) && buf[3] == ' ')
            resp = 1;
          else {
            if (buf[strlen(buf) - 1] == '\n')
              buf[strlen(buf) - 1] = 0;
            if (buf[strlen(buf) - 1] == '\r')
              buf[strlen(buf) - 1] = 0;
            if ((ptr = rindex(buf, '\n')) != NULL) {
              ptr++;
              if (isdigit((int)*ptr) && *(ptr + 3) == ' ')
                resp = 1;
            }
          }
        }
      } while (buf != NULL && resp == 0);

      if (buf == NULL)
        exit(-1);
     if (*ptr != '2') {
      if (hydra_send(sock, buffer2, strlen(buffer2), 0) < 0)
	exit(-1);
      resp = 0;
      do {
        ptr = buf = hydra_receive_line(sock);
        if (buf != NULL) {
          if (isdigit((int)buf[0]) && buf[3] == ' ')
            resp = 1;
          else {
            if (buf[strlen(buf) - 1] == '\n')
              buf[strlen(buf) - 1] = 0;
            if (buf[strlen(buf) - 1] == '\r')
              buf[strlen(buf) - 1] = 0;
            if ((ptr = rindex(buf, '\n')) != NULL) {
              ptr++;
              if (isdigit((int)*ptr) && *(ptr + 3) == ' ')
                resp = 1;
            }
          }
        }
      } while (buf != NULL && resp == 0);
      if (buf == NULL)
        exit(-1); 
      }

      if ((strstr(buf, "LOGIN") == NULL) && (strstr(buf, "NTLM") != NULL)) {
	smtp_auth_mechanism=AUTH_NTLM;
      }

#ifdef LIBOPENSSLNEW
      if ((strstr(buf, "LOGIN") == NULL) && (strstr(buf, "DIGEST-MD5") != NULL)) {
	smtp_auth_mechanism=AUTH_DIGESTMD5;
      }

      if ((strstr(buf, "LOGIN") == NULL) && (strstr(buf, "CRAM-MD5") != NULL)) {
	smtp_auth_mechanism=AUTH_CRAMMD5;
      }
#endif

      if ((strstr(buf, "LOGIN") == NULL) && (strstr(buf, "PLAIN") != NULL)) {
	smtp_auth_mechanism=AUTH_PLAIN;
      }

      if ((miscptr != NULL)&& (strlen(miscptr) > 0)) {
	      for (i = 0; i < strlen(miscptr); i++)
       		miscptr[i] = (char) toupper((int)miscptr[i]);

	       if (strncmp (miscptr,"LOGIN",5) == 0)
 		smtp_auth_mechanism=AUTH_LOGIN;

	       if (strncmp (miscptr,"PLAIN",5) == 0)
 		smtp_auth_mechanism=AUTH_PLAIN;

#ifdef LIBOPENSSLNEW
	       if (strncmp (miscptr,"CRAM-MD5",8) == 0)
 		smtp_auth_mechanism=AUTH_CRAMMD5;

	       if (strncmp (miscptr,"DIGEST-MD5",10) == 0)
 		smtp_auth_mechanism=AUTH_DIGESTMD5;
#endif

	       if (strncmp (miscptr,"NTLM",4) == 0)
 		smtp_auth_mechanism=AUTH_NTLM;

      }
      if (verbose) {
	switch (smtp_auth_mechanism) {
		case AUTH_LOGIN:
			hydra_report(stderr, "using SMTP LOGIN AUTH mechanism\n");
			break;
		case AUTH_PLAIN:
			hydra_report(stderr, "using SMTP PLAIN AUTH mechanism\n");
			break;
#ifdef LIBOPENSSLNEW
		case AUTH_CRAMMD5:
			hydra_report(stderr, "using SMTP CRAM-MD5 AUTH mechanism\n");
			break;
		case AUTH_DIGESTMD5:
			hydra_report(stderr, "using SMTP DIGEST-MD5 AUTH mechanism\n");
			break;
#endif
		case AUTH_NTLM:
			hydra_report(stderr, "using SMTP NTLM AUTH mechanism\n");
			break;
		}
      }
      free(buf);
      next_run = 2;
      break;
    case 2:			/* run the cracking function */
      next_run = start_smtpauth(sock, ip, port, options, miscptr, fp);
      break;
    case 3:			/* clean exit */
      if (sock >= 0) {
	sock = hydra_disconnect(sock);
      }
      hydra_child_exit(0);
      return;
    default:
      hydra_report(stderr, "Caught unknown return code, exiting!\n");
      hydra_child_exit(0);
      exit(-1);
    }
    run = next_run;
  }
}
